(* -*- sml-lex -*- *)

structure T = Tokens
structure E = ErrorMsg

type pos            = int
type svalue         = T.svalue
type ('a, 'b) token = ('a, 'b) T.token
type lexresult      = (svalue,pos) token

type lexarg         = { enterC : unit -> unit,
			leaveC : unit -> bool,
			handleEof : unit -> pos,
			newline : pos -> unit,
			error : pos * pos -> string -> unit }

type arg = lexarg

fun eof (arg: lexarg) = let
    val pos = #handleEof arg ()
in
    T.EOF (pos, pos)
end

local
    val idlist = [("class", T.CLASS),
		  ("extends", T.EXTENDS),
		  ("public", T.PUBLIC),
		  ("static", T.STATIC),
		  ("int", T.INT),
		  ("boolean", T.BOOLEAN),
		  ("void", T.VOID),
		  ("while", T.WHILE),
		  ("do", T.DO),
		  ("for", T.FOR),
		  ("if", T.IF),
		  ("else", T.ELSE),
		  ("return", T.RETURN),
		  ("break", T.BREAK),
		  ("continue", T.CONTINUE),
		  ("new", T.NEW),
		  ("true", T.TRUE),
		  ("false", T.FALSE),
		  ("this", T.THIS),
		  ("null", T.NULL)]
in
    fun idToken (t, p) =
	case List.find (fn (id, _) => id = t) idlist of
	    NONE => T.Id (Symbol.fromString t, p, p + size t)
	  | SOME (_, tok) => tok (p, p + size t)
end

%%
%s COMMENT;
%header (functor MiniJavaLexFun (structure Tokens: MiniJava_TOKENS));
%arg ({ enterC, leaveC, handleEof, newline, error });

letter=[a-zA-Z];
letdig=[a-zA-Z0-9_];
digit=[0-9];
white=[\ \t\r\f];


%%

<COMMENT>"/*"    =>  (enterC (); continue ());
<COMMENT>"*/"    =>  (if leaveC () then YYBEGIN INITIAL else (); continue ());
<COMMENT>"//.*\n" => (newline (yypos + size yytext - 1); continue ());
<COMMENT>"\n"    =>  (newline yypos; continue ());
<COMMENT>.       =>  (continue ());
<INITIAL>"/*"    => (YYBEGIN COMMENT; enterC (); continue ());
<INITIAL>"*/"    => (error (yypos, yypos + 2) "unmatched comment delimiter";
		     continue ());
<INITIAL>"//.*\n" => (newline (yypos + size yytext - 1); continue());
<INITIAL>{letter}{letdig}* => (idToken (yytext, yypos));
<INITIAL>"||"     =>(T.DBLBAR (yypos, yypos + 2));
<INITIAL>"&&"    => (T.DBLAMPERSAND (yypos, yypos + 2));
<INITIAL>">="    => (T.GREATEREQ (yypos, yypos + 2));
<INITIAL>">"     => (T.GREATERTHAN (yypos, yypos + 1));
<INITIAL>"<="    => (T.LESSEQ (yypos, yypos + 2));
<INITIAL>"<"     => (T.LESSTHAN (yypos, yypos + 1));
<INITIAL>"!="    => (T.NOTEQUALTO (yypos, yypos + 2));
<INITIAL>"=="    => (T.EQUALTO (yypos, yypos + 2));
<INITIAL>"="     => (T.EQUAL (yypos, yypos + 1));
<INITIAL>"*"     => (T.STAR (yypos, yypos + 1));
<INITIAL>"-"     => (T.MINUS (yypos, yypos + 1));
<INITIAL>"+"     => (T.PLUS (yypos, yypos + 1));
<INITIAL>"."     => (T.DOT (yypos, yypos + 1));
<INITIAL>"?"     => (T.QUESTION (yypos, yypos + 1));
<INITIAL>"{"     => (T.LCURL (yypos, yypos + 1));
<INITIAL>"}"     => (T.RCURL (yypos, yypos + 1));
<INITIAL>"["{white}*"]" => (T.LRSQUARE (yypos, yypos + size yytext));
<INITIAL>"["     => (T.LSQUARE (yypos, yypos + 1));
<INITIAL>"]"     => (T.RSQUARE (yypos, yypos + 1));
<INITIAL>"("     => (T.LPAR (yypos, yypos + 1));
<INITIAL>")"     => (T.RPAR (yypos, yypos + 1));
<INITIAL>";"     => (T.SEMI (yypos, yypos + 1));
<INITIAL>":"     => (T.COLON (yypos, yypos + 1));
<INITIAL>","     => (T.COMMA (yypos, yypos + 1));
<INITIAL>"!"     => (T.EXCLAM (yypos, yypos + 1));
<INITIAL>{digit}+ => (T.Integer
		      (valOf (LargeInt.fromString yytext)
		       handle _ => (error (yypos, yypos + size yytext) "number too large";
				    0),
		       yypos, yypos + size yytext));
<INITIAL>\n      => (newline yypos; continue ());
<INITIAL>{white} => (continue ());
<INITIAL>.       => (error (yypos, yypos)
		      ("illegal character " ^ yytext);
		      continue ());
