structure Semant : sig

    val semant : Ast.program * Source.inputSource ->
		 Label.label * Translate.frag list

end = struct

    structure T = Ast
    structure Ty = Types
    structure Sy = Symbol
    structure Tr = Translate

    structure M = Sy.Map

    type loopinfo = Label.label * Label.label
		    (* for continue and break (in that order) *)
    datatype labelinfo =
	LOOPLABEL of loopinfo
      | NONLOOPLABEL of Label.label		(* for break *)

    type varaccess =
	 Tr.lexp -> Tr.lexp (* takes access for "this" as its argument *)

    type env = (Ty.typ * varaccess) M.map
    type lenv = { current: loopinfo option, labels: labelinfo M.map }

    (* information about context that is constant across entire method *)
    type mcontext =
	 { mytyp: Ty.typ,
	   thisacc: Tr.lexp,
	   resinfo: (Ty.typ * Tr.lexp) option,
	   retlab: Label.label,
	   src: Source.inputSource }

    type context =
	 { env: env,
	   lenv: lenv,
	   region: T.region,
	   minfo: mcontext }

    fun simplectx src : context =
	{ env = M.empty,
	  lenv = { current = NONE, labels = M.empty },
	  region = (0, 0),
	  minfo = { mytyp = Ty.ERRORtyp,
		    thisacc = Tr.ldummy,
		    resinfo = NONE,
		    retlab = Label.new (),
		    src = src } }

    fun withMyTyp mytyp { env, lenv, region,
			  minfo = { mytyp = _, thisacc,
				    resinfo, retlab, src } } =
	{ env = env, lenv = lenv, region = region,
	  minfo = { src = src, retlab = retlab, thisacc = thisacc,
		    mytyp = mytyp, resinfo = resinfo } } : context

    fun withMinfo (thisacc, resinfo, retlab)
		  { env, lenv, region,
		    minfo = { mytyp, thisacc = _, resinfo = _,
			      retlab = _, src } } =
	{ env = env, lenv = lenv, region = region,
	  minfo = { src = src, retlab = retlab,
		    mytyp = mytyp, thisacc = thisacc,
		    resinfo = resinfo } } : context

    fun withRegion region { env, lenv, region = _, minfo } =
	{ env = env, lenv = lenv, region = region, minfo = minfo } : context

    fun withEnv env { env = _, lenv, region, minfo } = 
	{ env = env, lenv = lenv, region = region, minfo = minfo } : context

    fun withLoopLabel (label, linfo)
		      { env, lenv = { current, labels }, region, minfo } =
	{ minfo = minfo , env = env, region = region,
	  lenv = { current = current,
		   labels = M.insert (labels, label,
				      LOOPLABEL linfo) } } : context

    fun withNonLoopLabel (label, blab)
			 { env, lenv = { current, labels }, region, minfo } =
	{ env = env, region = region, minfo = minfo,
	  lenv = { current = current,
		   labels = M.insert (labels, label,
				      NONLOOPLABEL blab) } } : context

    fun loopctx ({ env, lenv = { current, labels }, region, minfo },
		 clab, blab) =
	{ env = env, region = region, minfo = minfo,
	  lenv = { current = SOME (clab, blab), labels = labels } }

    fun err { env, lenv, region,
	      minfo = { src, mytyp, thisacc, resinfo, retlab } } m =
	ErrorMsg.error src region ErrorMsg.COMPLAIN (concat m)
		       ErrorMsg.nullErrorBody

    fun bug m = raise Fail ("!!! IMPOSSIBLE: semant.sml: " ^ m)

    fun arrtyp Ty.ERRORtyp = Ty.ERRORtyp
      | arrtyp t = Ty.ARRtyp t

    fun isMyTyp (ctx: context) s =
	case #mytyp (#minfo ctx) of
	    Ty.CLASStyp (_, ref (Ty.CLASS c)) => s = #stamp c
	  | _ => false

    fun lub (Ty.INTtyp, Ty.INTtyp) = SOME Ty.INTtyp
      | lub (Ty.BOOLtyp, Ty.BOOLtyp) = SOME Ty.BOOLtyp
      | lub (Ty.ARRtyp t1, Ty.ARRtyp t2) =
	  (* notice: Java array-covariance bug! *)
	  Option.map Ty.ARRtyp (lub (t1, t2))
      | lub (Ty.NULLtyp, Ty.NULLtyp) = SOME Ty.NULLtyp
      | lub (Ty.NULLtyp, t as Ty.CLASStyp _) = SOME t
      | lub (t as Ty.CLASStyp _, Ty.NULLtyp) = SOME t
      | lub (Ty.NULLtyp, t as Ty.ARRtyp _) = SOME t
      | lub (t as Ty.ARRtyp _, Ty.NULLtyp) = SOME t
      | lub (t1 as Ty.CLASStyp (_, ref c1), t2 as Ty.CLASStyp (_, ref c2)) =
	  (case (c1, c2) of
	       (Ty.ERRORcls, _) => SOME t2
	     | (_, Ty.ERRORcls) => SOME t1
	     | (Ty.CLASS r1, Ty.CLASS r2) =>
	       let val d = Int.min (#depth r1, #depth r2)
		   val scl1 = List.drop (c1 :: #superclasses r1, #depth r1 - d)
		   val scl2 = List.drop (c2 :: #superclasses r2, #depth r2 - d)
		   fun search ([], []) = NONE
		     | search ((c as Ty.CLASS r) :: cs, Ty.CLASS r' :: c's) =
		         if #stamp r = #stamp r' then SOME c
			 else search (cs, c's)
		     | search (Ty.ERRORcls :: _, c :: _) = SOME c
		     | search (c :: _, Ty.ERRORcls :: _) = SOME c
		     | search _ = bug "lub"
		   fun mktyp NONE = NONE
		     | mktyp (SOME (c as Ty.CLASS { name, ... }))=
		         SOME (Ty.CLASStyp (name, ref c))
		     | mktyp (SOME Ty.ERRORcls) = SOME Ty.ERRORtyp
	       in
		   mktyp (search (scl1, scl2))
	       end)
      | lub (Ty.ERRORtyp, t2) = SOME t2
      | lub (t1, Ty.ERRORtyp) = SOME t1
      | lub _ = NONE

    fun lub' (t1, t2, ctx, m) =
	case lub (t1, t2) of
	    SOME t => t
	  | NONE => (err ctx ["types ", Ty.toString t1, " and ",
			      Ty.toString t2, " are not compatible (",
			      m, ")"];
		     Ty.ERRORtyp)

    fun insistLub args = ignore (lub' args)

    (* is t1 a supertype of t2? *)
    fun isSup (t1, t2) = let
	fun stampAtDepthIs (d, s, Ty.ERRORcls) = true
	  | stampAtDepthIs (d, s, Ty.CLASS { depth = d', stamp = s',
					     superclasses = scl, ... }) =
	    let fun stampOf (Ty.CLASS { stamp, ... }) = stamp
		  | stampOf Ty.ERRORcls = s
	    in
		(d = d' andalso s = s')
		orelse (d < d' andalso
			stampOf (List.nth (scl, d' - d - 1)) = s)
	    end
	fun check (Ty.INTtyp, Ty.INTtyp) = true
	  | check (Ty.BOOLtyp, Ty.BOOLtyp) = true
	  | check (Ty.ARRtyp t1, Ty.ARRtyp t2) =
	      (* notice: Java array-covariance bug! *)
	      check (t1, t2)
	  | check (Ty.ARRtyp _, Ty.NULLtyp) = true
	  | check (Ty.NULLtyp, Ty.NULLtyp) = true
	  | check (Ty.CLASStyp _, Ty.NULLtyp) = true
	  | check (Ty.CLASStyp (_, ref c1), Ty.CLASStyp (_, ref c2)) =
	      (case c1 of
		   Ty.ERRORcls => true
		 | Ty.CLASS { depth = d1, stamp = s1, ... } =>
		     stampAtDepthIs (d1, s1, c2))
	  | check (Ty.ERRORtyp, _) = true
	  | check (_, Ty.ERRORtyp) = true
	  | check _ = false
    in
	check (t1, t2)
    end

    fun insistSup (t1, t2, ctx, m) = let
	fun complain () =
	    err ctx [Ty.toString t1, " is not a supertype of ",
		     Ty.toString t2, "(", m, ")"]
    in
	if isSup (t1, t2) then () else complain ()
    end

    fun semant ({ main: T.mainclass, classes }, src) = let

	val ctx0 = simplectx src

	val classdeclmap =
	    let fun loop (m, [] : T.classdecl list) = m
		  | loop (m, d :: ds) =
		      if M.inDomain (m, #name d) then
			  (err (withRegion (#region d) ctx0)
			       ["duplicate declaration of class ",
				Sy.toString (#name d)];
			   loop (m, ds))
		      else
			  loop (M.insert (m, #name d, d), ds)
	    in
		loop (M.empty, classes)
	    end

	val classtypmap =
	    foldl (fn (d: T.classdecl, m) =>
		      M.insert (m, #name d, Ty.CLASStyp (#name d,
							 ref Ty.ERRORcls)))
	          M.empty
		  classes

	fun typ (T.NAMEDtyp sy, ctx) =
	      (case M.find (classtypmap, sy) of
		   SOME t => t
		 | NONE => (err ctx ["undefined class: ", Sy.toString sy];
			    Ty.ERRORtyp))
	  | typ (T.INTtyp, _) = Ty.INTtyp
	  | typ (T.BOOLtyp, _) = Ty.BOOLtyp
	  | typ (T.ARRtyp t, ctx) = arrtyp (typ (t, ctx))
	  | typ (T.MARKtyp (r, t), ctx) = typ (t, withRegion r ctx)


	(* For each class insert Ty.CLASS into classtypmap.
	 * This requires recursion into the inheritance/subclass
	 * hierarchy.  Detect cycles in hierarchy.
	 * Do not yet check consistency of method types as this
	 * requires the full hierarchy to be in place. *)
	fun initclasstypmap () = let

	    val nextstamp = ref 0
	    fun genstamp () =
		let val i = !nextstamp in nextstamp := i + 1; i end

	    val errorclasses = ref []

	    fun class subclasses (d: T.classdecl) = let
		val name = #name d
		fun err' m = err (withRegion (#region d) ctx0) m
		fun errorclass () =
		    (errorclasses := name :: !errorclasses;
		     Ty.ERRORcls)
		fun sameName n = Sy.eq (n, name)
	    in
		case M.find (classtypmap, #name d) of
		    SOME (Ty.CLASStyp (_, ref (c as Ty.CLASS _))) => c
		  | SOME (Ty.CLASStyp (_, r)) =>
		    if List.exists sameName subclasses then
			(err' ["cyclic inheritance involving ",
			       Sy.toString name];
			 errorclass ())
		    else if List.exists sameName (!errorclasses) then
			Ty.ERRORcls
		    else
			let val stamp = genstamp ()
			    val (depth, superclasses) =
				case #superclass d of
				    NONE => (0, [])
				  | SOME scsy =>
				      (case M.find (classdeclmap, scsy) of
					   NONE =>
					   (err' ["undefined (super-)class: ",
						  Sy.toString scsy];
					    (0, []))
					 | SOME scd =>
					   (case class (name :: subclasses) scd
					     of sc as Ty.CLASS { depth,
								 superclasses,
								 ... } =>
						(depth + 1,
						 sc :: superclasses)
					      | Ty.ERRORcls =>
						(0, [])))
			    val (instvars0, methods0, instvarenv0) =
				case superclasses of
				    Ty.CLASS  c :: _ =>
				      (#instvars c, #methods c, #instvarenv c)
				  | _ =>
				      (Ty.SlotMap.empty,
				       Ty.NameMap.empty,
				       Ty.NameMap.empty)
			    val (instvars, instvarenv) =
				let fun loop (_, []: T.instvardecl list,
					      vm, ve) =
					  (vm, ve)
				      | loop (k, vd :: vds, vm, ve) = let
					    val name = #name vd
					    val ctx =
						withRegion (#region vd) ctx0
					    val v = { name = name,
						      slot = k,
						      public = #public vd,
						      typ = typ (#typ vd, ctx)}
					in
					    if List.exists
						   (fn vd' =>
						       Sy.eq (#name vd, name))
						   vds then
						(err ctx
						     ["instance variable name clash: ",
						      Sy.toString name];
					         loop (k, vds, vm, ve))
					  else
					      loop (k + 1, vds,
						    Ty.SlotMap.insert
							(vm, k, v),
						    Ty.NameMap.insert
							(ve, name, v))
					end
				in
				    loop (Ty.SlotMap.numItems instvars0,
					  rev (#vardecls d),
					  instvars0, instvarenv0)
				end
			    val methods =
				let fun loop (_, [], mm) = mm
				      | loop (k, md :: mdl, mm) = let
					    val { public, restyp, name,
						  formals, body, region } = md
					    val ctx = withRegion region ctx0
					    fun procform ([], tl) = rev tl
					      | procform ((sy, t) :: fl, tl) =
						  if List.exists
							 (fn (sy', _) =>
							     Sy.eq (sy, sy'))
							 fl then
						      (err ctx ["duplicate formal parameter ",
								Sy.toString sy,
								" in method ",
								Sy.toString name];
						       procform (fl, tl))
						  else
						      procform
							  (fl,
							   typ (t, ctx) :: tl)
					    val restyp =
						case restyp of
						    NONE => NONE
						  | SOME t =>
						      SOME (typ (t, ctx))
					    val formals =
						procform (formals, [])
					    fun mk_m slot =
						{ slot = slot,
						  label = Label.new (),
						  formals = formals,
						  restyp = restyp,
						  public = public }
					    fun addto (k, slot) =
						loop (k, mdl,
						      Ty.NameMap.insert
							  (mm, name,
							   mk_m slot))
					in
					    if List.exists
						   (fn md' =>
						       Sy.eq (#name md', name))
						   mdl then
						(err ctx ["method name clash: ",
							  Sy.toString name];
						 loop (k, mdl, mm))
					    else
						case Ty.NameMap.find (mm, name)
						 of NONE => addto (k + 1, k)
						  | SOME m' =>
						    addto (k, #slot m')
					end
				in
				    loop (Ty.NameMap.numItems methods0,
					  rev (#methoddecls d), methods0)
				end
			    val c =
				Ty.CLASS { stamp = stamp,
					   depth = depth,
					   name = #name d,
					   superclasses = superclasses,
					   instvars = instvars,
					   instvarenv = instvarenv,
					   methods = methods }
			in
			    r := c;
			    c
			end
		  | _ => bug "initclasstypmap"
	    end
	in
	    M.app (ignore o class []) classdeclmap
	end

	val _ = initclasstypmap ()

	fun stripmarks (T.MARKexp (r, e), ctx) =
	      stripmarks (e, withRegion r ctx)
	  | stripmarks (e, ctx) = (e, ctx)

	fun exp (T.CONDexp (c, t, e), ctx) =
	    let val (c_ty, c_tr)  = exp (c, ctx)
		val (t_ty, t_tr) = exp (t, ctx)
		val (e_ty, e_tr) = exp (e, ctx)
	    in
		insistLub (c_ty, Ty.BOOLtyp, ctx, "conditional");
		(lub' (t_ty, e_ty, ctx, "conditional"),
		 Tr.cond (c_tr, t_tr, e_tr))
	    end
	  | exp (T.BINOPexp (bop as (T.LESSop | T.GREATERop |
				     T.LESSEQop | T.GREATEREQop),
			     e1, e2), ctx) =
	    let val (ty1, tr1) = exp (e1, ctx)
		val (ty2, tr2) = exp (e2, ctx)
	    in
		insistLub (ty1, Ty.INTtyp, ctx,
			   "left operand of " ^ T.binopName bop);
		insistLub (ty2, Ty.INTtyp, ctx,
			   "right operand of " ^ T.binopName bop);
		(Ty.BOOLtyp, Tr.compare (bop, tr1, tr2))
	    end
	  | exp (T.BINOPexp (bop as (T.EQUALop | T.NOTEQUALop),
			     e1, e2), ctx) =
	    let val (ty1, tr1) = exp (e1, ctx)
		val (ty2, tr2) = exp (e2, ctx)
	    in
		insistLub (ty1, ty2, ctx, "operands of " ^ T.binopName bop);
		(Ty.BOOLtyp, Tr.equal (bop, tr1, tr2))
	    end
	  | exp (T.BINOPexp (bop as (T.PLUSop | T.MINUSop | T.TIMESop),
			     e1, e2), ctx) =
	    let val (ty1, tr1) = exp (e1, ctx)
		val (ty2, tr2) = exp (e2, ctx)
	    in
		insistLub (ty1, Ty.INTtyp, ctx,
			   "left operand of " ^ T.binopName bop);
		insistLub (ty2, Ty.INTtyp, ctx,
			   "right operand of " ^ T.binopName bop);
		(Ty.INTtyp, Tr.arith (bop, tr1, tr2))
	    end
	  | exp (T.UOPexp (uop as T.NOTop, e), ctx) =
	    let val (ty, tr) = exp (e, ctx)
	    in
		insistLub (ty, Ty.BOOLtyp, ctx, "operand of " ^ T.uopName uop);
		(Ty.BOOLtyp, Tr.not tr)
	    end
	  | exp (T.UOPexp (uop as T.UMINUSop, e), ctx) =
	    let val (ty, tr) = exp (e, ctx)
	    in
		insistLub (ty, Ty.INTtyp, ctx, "operand of " ^ T.uopName uop);
		(Ty.INTtyp, Tr.neg tr)
	    end
	  | exp (T.CALLexp (e, el), ctx) =
	      (case stripmarks (e, ctx) of
		   (T.FETCHexp (T.SELlval (oe, sy)), ctx') =>
		     callexp (oe, sy, ctx', el, ctx)
		 | (T.FETCHexp (T.IDlval sy), ctx') =>
		     callexp (T.THISexp, sy, ctx', el, ctx)
		 | (_, ctx') => (err ctx' ["ill-formed method in method call"];
				 (Ty.ERRORtyp, Tr.dummy)))
	  | exp (T.INTexp i, ctx) = (Ty.INTtyp, Tr.int i)
	  | exp (T.BOOLexp b, ctx) = (Ty.BOOLtyp, Tr.bool b)
	  | exp (T.THISexp, ctx) =
	      (#mytyp (#minfo ctx), Tr.fetch (#thisacc (#minfo ctx)))
	  | exp (T.NULLexp, _) =
	      (Ty.NULLtyp, Tr.nilexp)
	  | exp (T.NEWOBJexp sy, ctx) =
	      let val ty = typ (T.NAMEDtyp sy, ctx)
	      in
		  (ty, Tr.newobj ty)
	      end
	  | exp (T.NEWARRexp (t, idxe), ctx) =
	      let val (idxty, idxtr) = exp (idxe, ctx)
		  val elemty = typ (t, ctx)
	      in
		  insistLub (idxty, Ty.INTtyp, ctx, "array length");
		  (arrtyp elemty, Tr.newarr idxtr)
	      end
	  | exp (T.FETCHexp lv, ctx) =
	    let val (ty, _, tr) = lvalue (lv, ctx)
	    in
		(ty, Tr.fetch tr)
	    end
	  | exp (T.ASSIGNexp (lv, e), ctx) =
	    let val (lty, writable, ltr) = lvalue (lv, ctx)
		val (ety, etr) = exp (e, ctx)
	    in
		insistSup (lty, ety, ctx, "operands in assignment");
		if writable then
		    (lty, Tr.assign (ltr, etr))
		else
		    (err ctx ["assignment to field declared final"];
		     (lty, Tr.dummy))
	    end
	  | exp (T.MARKexp (r, e), ctx) = exp (e, withRegion r ctx)

	and callexp (oe, sy, ctx', el, ctx: context) =
	    let fun checkargs ([], [], etrl) = rev etrl
		  | checkargs (fts as (_ :: _), [], etrl) =
		      (err ctx
			   ["insufficient number of arguments in method call"];
		       rev (map (fn _ => Tr.dummy) fts @ etrl))
		  | checkargs ([], _ :: _, etrl) =
		      (err ctx ["too many arguments in method call"];
		       rev etrl)
		  | checkargs (ft :: fts, e :: es, etrl) =
		      let val (ety, etr) = exp (e, ctx)
		      in
			  insistSup (ft, ety, ctx, "method argument");
			  checkargs (fts, es, etr :: etrl)
		      end
		fun look (mm, privacc, otr, c) =
		    case Ty.NameMap.find (mm, sy) of
			NONE => (err ctx' ["no such method: ", Sy.toString sy];
				 (Ty.ERRORtyp, Tr.dummy))
		      | SOME { public, formals, restyp, slot, label } =>
			  if public orelse privacc then
			      let val etrl = checkargs (formals, el, [])
				  val t =
				      case restyp of
					  NONE => Ty.VOIDtyp
					| SOME t => t
			      in
				  (t, Tr.call (otr, slot, etrl))
			      end
			  else
			      (err ctx' ["method has private access: ",
					 Sy.toString sy];
			       (Ty.ERRORtyp, Tr.dummy))
	    in
		case exp (oe, ctx') of
		    (Ty.CLASStyp (_, ref Ty.ERRORcls), _) =>
		      (Ty.ERRORtyp, Tr.dummy)
		  | (Ty.CLASStyp (_, ref (Ty.CLASS c)), otr) =>
		      if isMyTyp ctx (#stamp c) then
			  look (#methods c, true, otr, Ty.CLASS c)
		      else
			  look (#methods c, false, otr, Ty.CLASS c)
		  | (Ty.ERRORtyp, _) => (Ty.ERRORtyp, Tr.dummy)
		  | (t, _) => (err ctx' ["object type expected, found ",
					 Ty.toString t];
			       (Ty.ERRORtyp, Tr.dummy))
	    end

	and lvalue (T.IDlval sy, ctx) =
	      (case M.find (#env ctx, sy) of
		   SOME (ty, accf) => (ty, true, accf (#thisacc (#minfo ctx)))
		 | NONE => (err ctx ["undefined variable ", Sy.toString sy];
			    (Ty.ERRORtyp, true, Tr.ldummy)))
	  | lvalue (T.SUBlval (e1, e2), ctx) =
	    let val (ty1, tr1) = exp (e1, ctx)
		val (ty2, tr2) = exp (e2, ctx)
	    in
		insistLub (ty2, Ty.INTtyp, ctx, "subscript");
		case ty1 of
		    Ty.ARRtyp t => (t, true, Tr.sublval (tr1, tr2))
		  | Ty.ERRORtyp => (Ty.ERRORtyp, true, Tr.ldummy)
		  | _ => (err ctx ["array type expected, found ",
				   Ty.toString ty1];
			  (Ty.ERRORtyp, true, Tr.ldummy))
	    end
	  | lvalue (T.SELlval (e, sy), ctx) = let
		fun look (ve, privacc, otr, c) =
		    case Ty.NameMap.find (ve, sy) of
			NONE => (err ctx ["no such instance variable: ",
					  Sy.toString sy];
				 (Ty.ERRORtyp, true, Tr.ldummy))
		      | SOME { name, public, typ, slot } => 
			  if public orelse privacc then
			      (typ, true, Tr.sellval (otr, slot))
			  else
			      (err ctx
				   ["instance variable has private access: ",
				    Sy.toString sy];
			       (Ty.ERRORtyp, true, Tr.ldummy))
	    in
		case (exp (e, ctx), Sy.toString sy) of
		    ((Ty.CLASStyp (_, ref Ty.ERRORcls), _), _) =>
		      (Ty.ERRORtyp, true, Tr.ldummy)
		  | ((Ty.CLASStyp (_, ref (Ty.CLASS c)), otr), _) =>
		      if isMyTyp ctx (#stamp c) then
			  look (#instvarenv c, true, otr, Ty.CLASS c)
		      else
			  look (#instvarenv c, false, otr, Ty.CLASS c)
		 | ((Ty.ARRtyp _, otr), "length") =>
		     (Ty.INTtyp, false, Tr.arrlength otr)
		 | ((Ty.ERRORtyp, _), _) => (Ty.ERRORtyp, true, Tr.ldummy)
		 | ((t, _), _) => (err ctx ["object type expected, found ",
					    Ty.toString t];
				   (Ty.ERRORtyp, true, Tr.ldummy))
	    end

	datatype skind =
            LOOP of loopinfo * context -> Tr.exp
	  | NONLOOP of Label.label * context -> Tr.exp

	fun labeled (T.WHILEstm (e, s)) =
	      LOOP (fn ((clab, blab), ctx) => whilestm (e, s, clab, blab, ctx))
	  | labeled (T.DOstm (s, e)) =
	      LOOP (fn ((clab, blab), ctx) => dostm (s, e, clab, blab, ctx))
	  | labeled (T.FORstm (fi, c, ri, b)) =
	      LOOP (fn ((clab, blab), ctx) =>
		       forstm (fi, c, ri, b, clab, blab, ctx))
	  | labeled (T.MARKstm (r, s)) =
	      (case labeled s of
		   LOOP f =>
		     LOOP (fn (i, ctx) => f (i, withRegion r ctx))
		 | NONLOOP f =>
		     NONLOOP (fn (blab, ctx) => f (blab, withRegion r ctx)))
	  | labeled (T.LABELEDstm (l, s)) =
	      (case labeled s of
		   LOOP f =>
		     LOOP (fn (i, ctx) => f (i, withLoopLabel (l, i) ctx))
		 | NONLOOP f =>
		     NONLOOP (fn (blab, ctx) =>
				 f (blab, withNonLoopLabel (l, blab) ctx)))
	  | labeled s =
	      NONLOOP (fn (blab, ctx) => Tr.seq [stm (s, ctx), Tr.label blab])

	and whilestm (e, s, clab, blab, ctx) =
	    let val (ty, tr) = exp (e, ctx)
		val ctx' = loopctx (ctx, clab, blab)
	    in
		insistLub (ty, Ty.BOOLtyp, ctx,
			   "condition in WHILE statement");
		Tr.whilestm { clab = clab, blab = blab,
			      c = tr, b = stm (s, ctx') }
	    end

	and dostm (s, e, clab, blab, ctx) =
	    let val (ty, tr) = exp (e, ctx)
		val ctx' = loopctx (ctx, clab, blab)
	    in
		insistLub (ty, Ty.BOOLtyp, ctx, "condition in DO statement");
		Tr.dostm { clab = clab, blab = blab,
			   b = stm (s, ctx'), c = tr }
	    end

	and forstm (fi, c, ri, b, clab, blab, ctx) =
	    let fun explist (el, ctx) =
		    map (fn e => #2 (exp (e, ctx))) el
		val (itrl, ctx') =
		    case fi of
			T.EXPlist el => (explist (el, ctx), ctx)
		      | T.LOCVARdecls lvds => locvardecls (lvds, ctx)
		val (ty, tr) = exp (c, ctx')
		val ctx'' = loopctx (ctx', clab, blab)
	    in
		insistLub (ty, Ty.BOOLtyp, ctx', "condition in FOR statement");
		Tr.forstm { blab = blab, clab = clab,
			    init = itrl, c = tr,
			    rinit = explist (ri, ctx'),
			    b = stm (b, ctx'') }
	    end

	and stm (T.IGNstm e, ctx) = Tr.discard (#2 (exp (e, ctx)))
	  | stm (T.SEQstm (s1, s2), ctx) =
	      Tr.seq [stm (s1, ctx), stm (s2, ctx)]
	  | stm (T.IFstm (c, t, e), ctx) =
	    let val (cty, ctr) = exp (c, ctx)
	    in
		insistLub (cty, Ty.BOOLtyp, ctx, "condition in IF statement");
		Tr.ifstm (ctr, stm (t, ctx), stm (e, ctx))
	    end
	  | stm (T.WHILEstm (e, s), ctx) =
	      whilestm (e, s, Label.new (), Label.new (), ctx)
	  | stm (T.DOstm (s, e), ctx) =
	      dostm (s, e, Label.new (), Label.new (), ctx)
	  | stm (T.FORstm (fi, c, ri, b), ctx) =
	      forstm (fi, c, ri, b, Label.new (), Label.new (), ctx)
	  | stm (T.RETURNstm NONE, { minfo = { resinfo = NONE, retlab, ... }, ... }) =
	      Tr.jump retlab
	  | stm (T.RETURNstm (SOME e),
		 ctx as { minfo = { resinfo = NONE, retlab, ... }, ... }) =
	      (err ctx ["returning value from VOID method"];
	       ignore (exp (e, ctx));	(* check this anyway *)
	       Tr.jump retlab)
	  | stm (T.RETURNstm NONE,
		 ctx as { minfo = { resinfo = SOME _, retlab, ... }, ... }) =
	      (err ctx ["return with no value from non-VOID method"];
	       Tr.jump retlab)
	  | stm (T.RETURNstm (SOME e),
		 ctx as { minfo = { resinfo = SOME (rty, rta),
				    retlab, ... }, ... }) =
	    let val (ty, tr) = exp (e, ctx)
	    in
		insistSup (rty, ty, ctx, "return value");
		Tr.seq [Tr.assign (rta, tr), Tr.jump retlab]
	    end
	  | stm (T.NULLstm, _) = Tr.nullstm
	  | stm (T.LABELEDstm (l, s), ctx) =
	      (case labeled s of
		   LOOP f => let
		       val linfo = (Label.new (), Label.new ())
		       val ctx' = withLoopLabel (l, linfo) ctx
		   in
		       f (linfo, ctx')
		   end
		 | NONLOOP f => let
		       val blab = Label.new ()
		   in
		       f (blab, withNonLoopLabel (l, blab) ctx)
		   end)
	  | stm (T.BREAKstm NONE,
		 ctx as { lenv = { current = NONE, ... }, ... }) =
	      (err ctx ["break outside loop"];
	       Tr.dummy)
	  | stm (T.BREAKstm NONE,
		 { lenv = { current = SOME (clab, blab), ... }, ... }) =
	      Tr.jump blab
	  | stm (T.BREAKstm (SOME l), ctx as { lenv = { labels, ... }, ... }) =
	      (case M.find (labels, l) of
		   SOME (LOOPLABEL (clab, blab)) => Tr.jump blab
		 | SOME (NONLOOPLABEL blab) => Tr.jump blab
		 | NONE => (err ctx ["undefined label ", Sy.toString l];
			    Tr.dummy))
	  | stm (T.CONTINUEstm NONE,
		 ctx as { lenv = { current = NONE, ... }, ... }) =
	      (err ctx ["continue outside loop"];
	       Tr.dummy)
	  | stm (T.CONTINUEstm NONE,
		 { lenv = { current = SOME (clab, blab), ... }, ... }) =
	      Tr.jump clab
	  | stm (T.CONTINUEstm (SOME l),
		 ctx as { lenv = { labels, ... }, ... }) =
	      (case M.find (labels, l) of
		   SOME (LOOPLABEL (clab, blab)) => Tr.jump clab
		 | SOME (NONLOOPLABEL _) =>
		     (err ctx ["not a loop label: ", Sy.toString l];
		      Tr.dummy)
		 | NONE => (err ctx ["undefined label: ", Sy.toString l];
			    Tr.dummy))
	  | stm (T.LETstm (d, s), ctx) =
	      letstm (d, s, ctx)
	  | stm (T.MARKstm (r, s), ctx) =
	      stm (s, withRegion r ctx)

	and locvardecls (lvds, ctx) =
	    let fun one ({ name, typ = t, init, region }, (el, ctx)) =
		    let val ctx = withRegion region ctx
			val t = typ (t, ctx)
			val a = Tr.newtemp ()
			val env = M.insert (#env ctx, name, (t, fn _ => a))
			val ctx' = withEnv env ctx
			fun initall ([], el) = el
			  | initall ((ty, lval, T.EXPinit e) :: rest, el) =
			    let val (ty', tr) = exp (e, ctx)
			    in
				insistSup (ty, ty', ctx, "initializer");
				initall (rest, Tr.assign (lval, tr) :: el)
			    end
			  | initall ((Ty.ARRtyp elty, lval, T.ARRinit il)
				     :: rest, el) =
			    let val alval = Tr.newtemp ()
				val aval = Tr.fetch alval
				val atr =
				    Tr.newarr (Tr.int (LargeInt.fromInt
							   (length il)))
				val aalloc =
				    Tr.discard (Tr.assign (alval, atr))
				val ainit = Tr.discard (Tr.assign (lval, aval))
				fun build (_, []) = rest
				  | build (k, i :: il) =
				      (elty,
				       Tr.sublval (aval, Tr.int k), i)
				      :: build (k + 1, il)
			    in
				initall (build (0, il), ainit :: aalloc :: el)
			    end
			  | initall ((_, _, T.ARRinit _) :: rest, el) =
			      (err ctx ["unexpected array initializer"];
			       initall (rest, el))
		    in
			case init of
			    NONE => (el, ctx')
			  | SOME i => (initall ([(t, a, i)], el), ctx')
		    end
		val (el, ctx'') = foldl one ([], ctx) lvds
	    in
		(rev el, ctx'')
	    end

	and letstm (lvds, s, ctx) =
	    let val (el, ctx') = locvardecls (lvds, ctx)
	    in
		Tr.seq [Tr.seq el, stm (s, ctx')]
	    end

	fun isSupMeth (m: Ty.method, m': Ty.method) =
	    let fun resIsSup (NONE, NONE)  = true
		  | resIsSup (SOME _, NONE) = false
		  | resIsSup (NONE, SOME _) = false
		  | resIsSup (SOME t, SOME t') =
		    isSup (t, t')
	    in
		#public m = #public m' andalso
		(* covariant: *)
		resIsSup (#restyp m, #restyp m') andalso
		(* contravariant: *)
		ListPair.allEq isSup (#formals m', #formals m)
	    end

	fun checkclass { name = cname,
			 superclass, vardecls, methoddecls, region } =
	    case M.find (classtypmap, cname) of
		SOME (mytyp as Ty.CLASStyp (_, ref (Ty.CLASS c))) =>
		  let val ctx = withMyTyp mytyp (withRegion region ctx0)
		      fun onevar ({ name, typ, public, slot}, env) =
			  M.insert (env, name,
				    (typ, fn a =>
					     Tr.sellval (Tr.fetch a, slot)))
		      val env = Ty.NameMap.foldr onevar M.empty (#instvarenv c)
		      val ctx = withEnv env ctx
		      val methods = #methods c
		      val supermethods =
			  case #superclasses c of
			      Ty.CLASS sc :: _ => #methods sc
			    | _ => Ty.NameMap.empty
		      fun onemethod { public, restyp, name, formals,
				      body, region } =
			  let val m =
				  case Ty.NameMap.find (methods, name) of
				      NONE => bug "method"
				    | SOME m =>
				        (case Ty.NameMap.find
						  (supermethods, name) of
					     SOME sm =>
					       if isSupMeth (sm, m) then ()
					       else err ctx
							["incompatible type of overriding method ",
							 Sy.toString name,
							 " in class ",
							 Sy.toString cname]
					   | NONE => ();
				         m)
			      val frame =
				  Tr.frame (List.length formals, #restyp m)
			      fun foldl3 f i (x :: xs, y :: ys, z :: zs) =
				    foldl3 f (f (x, y, z, i)) (xs, ys, zs)
				| foldl3 _ i _ = i
			      val resinfo = Tr.resinfo frame
			      val ctx =
				  withMinfo (Tr.thisacc frame,
					     resinfo, Tr.retlab frame)
					    (withRegion region ctx)
			      fun oneformal ((fname, _), t, acc, env) =
				  M.insert (env, fname, (t, fn _ => acc))
			      val env = foldl3 oneformal
					       (#env ctx)
					       (formals, #formals m,
						Tr.argaccl frame)
			      val ctx = withEnv env ctx
			      val entrylab = #label m
			  in
			      Tr.codefrag (frame, stm (body, ctx))
			  end
		      val methodfrags = map onemethod methoddecls
		      val classfrag = Tr.classfrag c
		  in
		      classfrag :: methodfrags
		  end
	      | SOME _ => []		(* previously found to be an error *)
	      | NONE => bug "checkmethod"
	val frags = M.foldl (fn (c, l) => checkclass c @ l) [] classdeclmap
	val mainframe = Tr.main ()
	val mainfrag =
	    Tr.codefrag (mainframe, stm (#body main,
					 withRegion (#region main) ctx0))
    in
	(Tr.entrylab mainframe, mainfrag :: frags)
    end
end
